/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									     *
 * Library:	libsubdiv2.5D						     *
 *									     *
 * Module:	subdiv2.5D_gfx.c					     *
 *									     *
 * Description:	graphics functions					     *
 *									     *
 * Date:	20000818						     *
 *									     *
 * Author:	Francisco MORN BURGOS   (http://www.gti.ssr.upm.es./~fmb/)  *
 *		Grupo Tratamiento Imgenes, Univ. Politcnica Madrid, Spain  *
 *									     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/* You must have (compiled and) installed in your system the GLUT (OpenGL    */
/* Utility Toolkit), which will in turn use OpenGL's GL and GLU libraries.   */
/* The latter two can be the native OpenGL libraries, if they came with your */
/* OS, or the ones from the public Mesa 3D Graphics Library, if you've also  */
/* (compiled and) installed it in your system.				     */
#include <GL/glut.h>
#include <math.h>

#ifndef WIN32
#  include <unistd.h>
#  define SLEEP(n)  sleep(n)		/* in UNIX, it's "sleep(s)" */
#else
#  define SLEEP(n)  Sleep(n*1000)	/* in MS-Windows, it's "Sleep(ms)" */
#endif

#include "subdiv2.5D_priv.h"



#define RED(r)	   r, 0.0F, 0.0F
#define GRN(g)	0.0F,    g, 0.0F
#define BLU(b)	0.0F, 0.0F,    b
#define CYN(c)	0.0F,    c,    c
#define MAG(m)	   m, 0.0F,    m
#define YEL(y)	   y,    y, 0.0F
#define GRY(g)	   g,    g,    g
#define BLACK	GRY(0.0F)
#define WHITE	GRY(1.0F)

/* display lists (1 Kthanx, mar...) */
#define N_DLS		2
#define DL_MAIN_MESH	0
#define DL_TGT_MESH	1
#define TM		sdgMesh[DL_MAIN_MESH]
static sdMeshPtr	sdgMesh[N_DLS]		= {NULL, NULL};
static GLenum		sdgPolyMode[N_DLS][2]	= {{GL_FILL, GL_FILL},
						   {GL_LINE, GL_LINE}};
static float		sdgDiffuse[N_DLS][2][4]	= {{{0.2F, 1.0F, 0.8F, 1.0F},
						    {1.0F, 0.8F, 0.2F, 1.0F}},
						   {{0.7F, 0.7F, 0.2F, 1.0F},
						    {0.7F, 0.2F, 0.7F, 1.0F}}},
			sdgSpecular[N_DLS][4]	= {{0.5F, 0.1F, 0.5F, 1.0F},
						   {0.3F, 0.3F, 0.0F, 1.0F}},
			sdgShininess[N_DLS][1]	= {{88.0F},
						   {88.0F}},
			sdgWhite[4]		= {WHITE, 1.0F},
			sdgBlack[4]		= {BLACK, 1.0F},
			sdgRGB[3][4]		= {{RED(1.0F), 1.0F},
						   {GRN(1.0F), 1.0F},
						   {BLU(1.0F), 1.0F}},
			sdgCMY[3][4]		= {{CYN(1.0F), 1.0F},
						   {MAG(1.0F), 1.0F},
						   {YEL(1.0F), 1.0F}};

static GLenum sdgGLmode = GL_RENDER;

static sdVector sdgPickedVertexInitPos = {0.0F, 0.0F, 0.0F};



typedef enum {CAM_CAR_ABS, CAM_CAR_REL, CAM_SPH_REL, AIM_CAR_REL} sdgCoordKind;

static void sdgSetCamera(sdgCoordKind ck, float c1, float c2, float c3)
{
  static sdVector aim = {0.0F, 0.0F, 0.0F},
		  cam = {0.0F, 0.0F, 1.0F},
		  upv = {0.0F, 1.0F, 0.0F};
  static sdBoolean inc_phi = TRUE;
  float D, theta, phi,
	lightPos[4] = {0.0F, 1.3F, 0.0F, 1.0F};	/* positional, above object */

  switch (ck) {
    case CAM_CAR_ABS:
      cam.x = c1; cam.y = c2; cam.z = c3;
      break;
    case CAM_CAR_REL:
      cam.x += c1; cam.y += c2; cam.z += c3;
      break;
    case CAM_SPH_REL:
      D = (float) sqrt(cam.x*cam.x + cam.y*cam.y + cam.z*cam.z) + c1;
      theta = (float) atan2(cam.x, cam.z) + c2;
      phi = (float) atan2(cam.y, sqrt(cam.x*cam.x+cam.z*cam.z)) +
	    (inc_phi? +c3 : -c3);
      if (phi < -M_PI_2) {
	phi = -M_PI - phi; inc_phi = !inc_phi; theta += M_PI; upv.y = -upv.y;
      } else if (phi > +M_PI_2) {
	phi = +M_PI - phi; inc_phi = !inc_phi; theta += M_PI; upv.y = -upv.y;
      }
      cam.x = (float) (D * cos(phi) * sin(theta));
      cam.y = (float) (D * sin(phi));
      cam.z = (float) (D * cos(phi) * cos(theta));
      break;
    case AIM_CAR_REL:
      aim.x += c1; aim.y += c2; aim.z += c3;
      break;
  }
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(cam.x, cam.y, cam.z, aim.x, aim.y, aim.z, upv.x, upv.y, upv.z);
  glLightfv(GL_LIGHT2, GL_POSITION, lightPos);
} /* sdgSetCamera */


static void sdgRotateCameraLeft(void)
{
  sdgSetCamera(CAM_SPH_REL, 0, -8 DEG, 0);
} /* sdgRotateCameraLeft */


static void sdgRotateCameraRight(void)
{
  sdgSetCamera(CAM_SPH_REL, 0, +8 DEG, 0);
} /* sdgRotateCameraRight */


static void sdgRotateCameraUp(void)
{
  sdgSetCamera(CAM_SPH_REL, 0, 0, +8 DEG);
} /* sdgRotateCameraUp */


static void sdgRotateCameraDown(void)
{
  sdgSetCamera(CAM_SPH_REL, 0, 0, -8 DEG);
} /* sdgRotateCameraDown */


static void sdgMoveCameraLeft(void)
{
  sdgSetCamera(AIM_CAR_REL, -0.17F, 0, 0);
} /* sdgMoveCameraLeft */


static void sdgMoveCameraRight(void)
{
  sdgSetCamera(AIM_CAR_REL, +0.17F, 0, 0);
} /* sdgMoveCameraRight */


static void sdgMoveCameraUp(void)
{
  sdgSetCamera(AIM_CAR_REL, 0, +0.17F, 0);
} /* sdgMoveCameraUp */


static void sdgMoveCameraDown(void)
{
  sdgSetCamera(AIM_CAR_REL, 0, -0.17F, 0);
} /* sdgMoveCameraDown */


static void sdgMoveCameraForward(void)
{
  sdgSetCamera(CAM_SPH_REL, -1, 0, 0);
} /* sdgMoveCameraForward */


static void sdgMoveCameraBackward(void)
{
  sdgSetCamera(CAM_SPH_REL, +1, 0, 0);
} /* sdgMoveCameraBackward */


static void sdgReshapeWindow(int w, int h)
{
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(10.0, (GLdouble)w/h, 1.0, 23.0);
} /* sdgReshapeWindow */


static void sdgInitDisplay(int width, int height, char *title)
{
  float lightAmb[3][4] = {{GRY(0.5F), 1.0F},
			  {BLACK,     1.0F},
			  {BLACK,     1.0F}},
	lightDif[3][4] = {{GRY(0.5F), 1.0F},
			  {RED(0.7F), 1.0F},
			  {BLU(0.5F), 1.0F}},
	lightPos[3][4] = {{ 0.0F, 0.0F, 1.0F, 0.0F},
			  {-1.0F, 0.0F, 0.0F, 0.0F},
			  { 0.0F, 0.0F, 0.0F, 1.0F}};

  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(width, height);
  glutCreateWindow(title);
  glClearColor(MAG(0.1F), 0.0);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);	/* for highlighting */
  if (TM->closed)
    glEnable(GL_CULL_FACE);
  else
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, TRUE);
  glLightfv(GL_LIGHT0, GL_AMBIENT,  lightAmb[0]);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  lightDif[0]);
  glLightfv(GL_LIGHT0, GL_POSITION, lightPos[0]);
  glLightfv(GL_LIGHT1, GL_AMBIENT,  lightAmb[1]);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  lightDif[1]);
  glLightfv(GL_LIGHT1, GL_POSITION, lightPos[1]);
  glLightfv(GL_LIGHT2, GL_AMBIENT,  lightAmb[2]);
  glLightfv(GL_LIGHT2, GL_DIFFUSE,  lightDif[2]);
  glLightfv(GL_LIGHT2, GL_POSITION, lightPos[2]);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  sdgSetCamera(CAM_CAR_ABS, 2.3F, 8.8F, 6.9F);
  glutReshapeFunc(sdgReshapeWindow);
} /* sdgInitDisplay */


static void sdgDestroyMeshAndCloseDisplay(void)
{
  sdDestroyMesh(&TM);
  exit(0);
} /* sdgDestroyMeshAndCloseDisplay */


static void sdgDisplayTriangleAtLast(sdTrianglePtr t)
{
  int k;
  sdPointPtr vtx;

  if (TM->normalsPerVertex) {
    for (k=0; k<3; k++) {
      vtx = t->vtx[k];
      glNormal3f(vtx->normal.x, vtx->normal.y, vtx->normal.z);
      glVertex3f(vtx->pos.x, vtx->pos.y, vtx->pos.z);
    }
  } else {
    glNormal3f(t->normal.x, t->normal.y, t->normal.z);
    for (k=0; k<3; k++) {
      vtx = t->vtx[k];
      glVertex3f(vtx->pos.x, vtx->pos.y, vtx->pos.z);
    }
  }
} /* sdgDisplayTriangleAtLast */


static void sdgDisplayTriangle(sdTrianglePtr t)
{
  int i, j, k, Nmidpoints, splitedge[3];
  sdPointPtr p[3], mp[3];
  sdTrianglePtr ta;
  sdTriangle taux;
  float d1, d2;

  if (!t)
    return;

  if (t->subdiv)
    for (k=0; k<4; k++)
      sdgDisplayTriangle(t->son[k]);
  else {
    for (k=0; k<3; k++)
      p[k] = t->vtx[k];
    if (sdgGLmode == GL_SELECT) {
      glLoadName(t->n);
      glBegin(GL_TRIANGLES);
      for (k=0; k<3; k++)
	glVertex3f(p[k]->pos.x, p[k]->pos.y, p[k]->pos.z);
      glEnd();
    } else if (sdgGLmode == GL_RENDER) {
      if (!TM->conforming) {	/* don't worry about the eventual cracks */
	sdgDisplayTriangleAtLast(t);
      } else {			/* DO worry about the eventual cracks */
	for (k=Nmidpoints=0; k<3; k++)
	  if ((ta = t->adt[k]) && ta->subdiv) {
	    splitedge[Nmidpoints++] = k;
	    i = IDX_VTX_IN_T(p[k], ta);
	    mp[k] = ta->son[i]->vtx[sdM1[i]];
	  }
	switch (Nmidpoints) {
	  case 0:	/* no cracks after all => only one triangle is needed */
	    sdgDisplayTriangleAtLast(t);
	    break;
	  case 1:	/* two triangles are needed to eliminate the crack */
	    i = splitedge[0];
	    taux.vtx[0] = p[i];
	    taux.vtx[1] = mp[i];
	    taux.vtx[2] = p[sdM1[i]];
	    sdmFillTriangleNormalAndK(&taux, TRUE);
	    sdgDisplayTriangleAtLast(&taux);
	    taux.vtx[0] = p[sdM1[i]];
	    taux.vtx[1] = mp[i];
	    taux.vtx[2] = p[sdP1[i]];
	    sdmFillTriangleNormalAndK(&taux, TRUE);
	    sdgDisplayTriangleAtLast(&taux);
	    break;
	  case 2:	/* three triangles are needed to eliminate the cracks */
	    i = splitedge[0]; j = splitedge[1];
	    if (j == i+1) {	/* (i, j) = (0, 1) or (1, 2) */
	      taux.vtx[0] = mp[j];
	      taux.vtx[1] = mp[i];
	      taux.vtx[2] = p[sdP1[i]];
	      sdmFillTriangleNormalAndK(&taux, TRUE);
	      sdgDisplayTriangleAtLast(&taux);
	      d1 = sdmVecNorm(sdmVecDiff(&p[sdM1[i]]->pos, &mp[i]->pos));
	      d2 = sdmVecNorm(sdmVecDiff(&p[i]->pos, &mp[j]->pos));
	      if (d1 < d2) {
		taux.vtx[0] = p[i];
		taux.vtx[1] = mp[i];
		taux.vtx[2] = p[sdM1[i]];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
		taux.vtx[0] = p[sdM1[i]];
		taux.vtx[1] = mp[i];
		taux.vtx[2] = mp[j];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
	      } else {
		taux.vtx[0] = p[i];
		taux.vtx[1] = mp[i];
		taux.vtx[2] = mp[j];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
		taux.vtx[0] = mp[j];
		taux.vtx[1] = p[sdM1[i]];
		taux.vtx[2] = p[i];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
	      }
	    } else {		/* (i, j) = (0, 2) */
	      taux.vtx[0] = p[i];
	      taux.vtx[1] = mp[i];
	      taux.vtx[2] = mp[j];
	      sdmFillTriangleNormalAndK(&taux, TRUE);
	      sdgDisplayTriangleAtLast(&taux);
	      d1 = sdmVecNorm(sdmVecDiff(&p[sdP1[i]]->pos, &mp[j]->pos));
	      d2 = sdmVecNorm(sdmVecDiff(&p[sdM1[i]]->pos, &mp[i]->pos));
	      if (d1 < d2) {
		taux.vtx[0] = mp[j];
		taux.vtx[1] = mp[i];
		taux.vtx[2] = p[sdP1[i]];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
		taux.vtx[0] = p[sdP1[i]];
		taux.vtx[1] = p[sdM1[i]];
		taux.vtx[2] = mp[j];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
	      } else {
		taux.vtx[0] = mp[j];
		taux.vtx[1] = mp[i];
		taux.vtx[2] = p[sdM1[i]];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
		taux.vtx[0] = p[sdM1[i]];
		taux.vtx[1] = mp[i];
		taux.vtx[2] = p[sdP1[i]];
		sdmFillTriangleNormalAndK(&taux, TRUE);
		sdgDisplayTriangleAtLast(&taux);
	      }
	    }
	    break;
	  case 3:	/* four triangles are needed to eliminate the cracks */
	    for (i=0; i<3; i++) {
	      taux.vtx[0] = p[i];
	      taux.vtx[1] = mp[i];
	      taux.vtx[2] = mp[sdM1[i]];
	      sdmFillTriangleNormalAndK(&taux, TRUE);
	      sdgDisplayTriangleAtLast(&taux);
	    }
	    taux.vtx[0] = mp[0];
	    taux.vtx[1] = mp[1];
	    taux.vtx[2] = mp[2];
	    sdmFillTriangleNormalAndK(&taux, TRUE);
	    sdgDisplayTriangleAtLast(&taux);
	    break;
	}
      }
    }
  }
} /* sdgDisplayTriangle */


static void sdgCreateDisplayList(int dl)
{
  int i;

  if (!sdgMesh[dl])
    return;

  glDeleteLists(dl+1, 1);
  glNewList(dl+1, GL_COMPILE);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, sdgDiffuse[dl][0]);
  glMaterialfv(GL_BACK,  GL_DIFFUSE, sdgDiffuse[dl][1]);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,  sdgSpecular[dl]);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, sdgShininess[dl]);
  glPolygonMode(GL_FRONT, sdgPolyMode[dl][0]);
  glPolygonMode(GL_BACK,  sdgPolyMode[dl][1]);
  glBegin(GL_TRIANGLES);
  for (i=0; i<sdgMesh[dl]->NmotherTriangles; i++)
    sdgDisplayTriangle(sdgMesh[dl]->motherTriangle[i]);
  glEnd();
  glEndList();
  sdgMesh[dl]->refresh = FALSE;
} /* sdgCreateDisplayList */


static void sdgDisplayMesh(void)
{
  int dl, eno;
  sdTriangleLst tl;
  sdVectorLst path, step;
  sdVectorPtr v0, v1;
  sdEdgePtr e;
  float x, y, z;

  if (!TM)
    return;

  if (TM->verbose > 1)
    fprintf(stderr,
	    "Displaying  %6d leaf (%6d total) triangles and %6d vertices...  ",
	    TM->NleafTriangles, TM->Ntriangles, TM->Npoints);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  for (dl=0; dl<N_DLS; dl++)
    if (sdgMesh[dl] && sdgMesh[dl]->display) {
      if (sdgMesh[dl]->refresh)
	sdgCreateDisplayList(dl);
      glCallList(dl+1);
      if (sdgMesh[dl]->tmpTriangleList) {
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glBegin(GL_TRIANGLES);
	for (tl=sdgMesh[dl]->tmpTriangleList; tl; tl=tl->next)
	  sdgDisplayTriangle(tl->t);
	glEnd();
      }
      if (sdgMesh[dl]->targetMeshPath4edge)
	for (eno=0; eno<sdgMesh[dl]->Nedges; eno++) {
	  e = sdgMesh[dl]->edge + eno;
	  if (e->highlighted && (path=sdgMesh[dl]->targetMeshPath4edge[eno])) {
	    v0 = &e->vtx[0]->pos;  v1 = &e->vtx[1]->pos;
	    glLineWidth(2.0F);
	    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, sdgRGB[1]);
	    glBegin(GL_LINES);
	    glVertex3f(v0->x, v0->y, v0->z);
	    glVertex3f(v1->x, v1->y, v1->z);
	    glEnd();
	    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, sdgWhite);
	    glBegin(GL_LINE_STRIP);
	    for (step=path; step; step=step->next) {
	      glVertex3f(x = step->pos.x, y = step->pos.y, z = step->pos.z);
	      if (!step->next && (x != v1->x || y != v1->y || z != v1->z))
		fprintf(stderr, "\n  Incomplete path!  ");
	    }
	    glEnd();
	    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, sdgBlack);
	    glLineWidth(1.0F);
	  }
	}
    }
  glutSwapBuffers();
  if (TM->verbose > 1)
    fprintf(stderr, "Done.\n");
} /* sdgDisplayMesh */


void sdgDisplayMainMesh(void)
{
  sdgDisplayMesh();
} /* sdgDisplayMainMesh */


static void sdgSubdivAllLeafTrianglesOnce(void)
{
  sdSubdivAllLeafTrianglesOnce(TM);
} /* sdgSubdivAllLeafTrianglesOnce */


static void sdgSubdivAllLeafTrianglesUntilSmooth(void)
{
  sdSubdivAllLeafTrianglesUntilSmooth(TM);
} /* sdgSubdivAllLeafTrianglesUntilSmooth */


static void sdgMergeAllLeafTrianglesOnce(void)
{
  sdMergeAllLeafTrianglesOnce(TM);
} /* sdgMergeAllLeafTrianglesOnce */


static void sdgHighlightNextEdgeTargetMeshPath(void)
{
  static int e = -1;

  if (!TM->edge)
    return;

  if (++e == TM->Nedges) {
    e = -1;
    TM->edge[TM->Nedges-1].highlighted = FALSE;
  } else {
    TM->edge[e].highlighted = TRUE;
    if (e > 0)
      TM->edge[e-1].highlighted = FALSE;
  }
} /* sdgHighlightNextEdgeTargetMeshPath */


static void sdgHighlightTriangle(sdTrianglePtr t, int seconds)
{
  int k;

  if (!t)
    return;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glCallList(DL_MAIN_MESH+1);
  glBegin(GL_TRIANGLES);
  for (k=0; k<3; k++) {
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, sdgRGB[k]);
    sdgDisplayTriangle(t->adt[k]);
  }
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, sdgWhite);
  sdgDisplayTriangle(t);
  glEnd();
  glutSwapBuffers();
  SLEEP(seconds);
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, sdgBlack);
} /* sdgHighlightTriangle */


#define BUFSIZE	512

static sdTrianglePtr sdgPickTriangle(int x, int y)
{
  sdTrianglePtr t;
  GLuint selectBuf[BUFSIZE], *ptr, zmin, closest;
  GLint viewport[4], i, hits;

  /* create a 3x3 pixel picking region around cursor and pick triangle(s) */
  glGetIntegerv(GL_VIEWPORT, viewport);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluPickMatrix(x, viewport[3] - y, 5.0, 5.0, viewport);
  gluPerspective(10.0, (GLdouble) viewport[2] / viewport[3], 1.0, 23.0);
  glSelectBuffer(BUFSIZE, selectBuf);
  glRenderMode(sdgGLmode = GL_SELECT);
  glInitNames();
  glPushName(~0);
  for (i=0; i<TM->NmotherTriangles; i++)
    sdgDisplayTriangle(TM->motherTriangle[i]);
  glFlush();
  glPopMatrix();

  /* process hit(s) */
  hits = glRenderMode(sdgGLmode = GL_RENDER);
  for (i=0, ptr=selectBuf+2, zmin=closest=~0; i<hits; i++, ptr+=4)
    if (*ptr < zmin) {
      zmin = *ptr;
      closest = *(ptr+1);
    }
  if (closest == ~0 || !(t = sdlFindLeafTriangleByID(TM, closest))) {
    BEEP;
    return NULL;
  }
  return t;
} /* sdgPickTriangle */


static void sdgMoveVertex(int x, int y)
{
  GLdouble mdlM[16], prjM[16], xw, yw, zw, x3D, y3D, z3D;
  GLint viewport[4], d0;

  if (!TM->pickedVertex)
    return;

  glGetDoublev(GL_MODELVIEW_MATRIX, mdlM);
  glGetDoublev(GL_PROJECTION_MATRIX, prjM);
  glGetIntegerv(GL_VIEWPORT, viewport);
  gluProject(TM->pickedVertex->pos.x, TM->pickedVertex->pos.y,
	     TM->pickedVertex->pos.z, mdlM, prjM, viewport, &xw, &yw, &zw);
  gluUnProject(x, viewport[3] - y, zw, mdlM, prjM, viewport, &x3D, &y3D, &z3D);
  TM->pickedVertex->pos.x = (float) x3D;
  TM->pickedVertex->pos.y = (float) y3D;
  TM->pickedVertex->pos.z = (float) z3D;
  if ((d0 = sdlVertexDepth(TM, TM->pickedVertex) - 1) < 0) d0 = 0;
  sdRecalculateAllVertexPositions(TM, d0, TM->depth, NBPDN, FALSE);
  glutPostRedisplay();
} /* sdgMoveVertex */


static void sdgUndoMoveVertex(void)
{
  int d0;

  TM->pickedVertex->pos = sdgPickedVertexInitPos;
  if ((d0 = sdlVertexDepth(TM, TM->pickedVertex) - 1) < 0) d0 = 0;
  sdRecalculateAllVertexPositions(TM, d0, TM->depth, NBPDN, FALSE);
} /* sdgUndoMoveVertex */


static void sdgPickVertex(int x, int y)
{
  GLuint selectBuf[BUFSIZE], *ptr, zmin, closest;
  GLint viewport[4], i, hits;
  sdPointPtr pt;

  /* create a 3x3 pixel picking region around cursor and pick vertex(es) */
  glGetIntegerv(GL_VIEWPORT, viewport);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluPickMatrix(x, viewport[3] - y, 3.0, 3.0, viewport);
  gluPerspective(10.0, (GLdouble) viewport[2] / viewport[3], 1.0, 23.0);
  glSelectBuffer(BUFSIZE, selectBuf);
  glRenderMode(sdgGLmode = GL_SELECT);
  glInitNames();
  glPushName(~0);
  for (pt=TM->firstVertex; pt; pt=pt->next) {
    glLoadName(ABS(pt->n));
    glBegin(GL_POINTS);
    glVertex3f(pt->pos.x, pt->pos.y, pt->pos.z);
    glEnd();
  }
  glFlush();
  glPopMatrix();

  /* process hit(s) */
  hits = glRenderMode(sdgGLmode = GL_RENDER);
  for (i=0, ptr=selectBuf+2, zmin=closest=~0; i<hits; i++, ptr+=4)
    if (*ptr < zmin) {
      zmin = *ptr;
      closest = *(ptr+1);
    }
  if (closest == ~0 || !(pt = sdlFindVertexByID(TM, closest))) {
    BEEP;
    TM->pickedVertex = NULL;
  } else {
    sdgPickedVertexInitPos = pt->pos;
    TM->pickedVertex = pt;
    if (TM->verbose)
      sdiPrintPoint(TM, pt, USELESS, FALSE);
    sdgMoveVertex(x, y);
  }
} /* sdgPickVertex */


static void sdgPickTriangleOrVertex(int x, int y)
{
  if (TM->progressive)
    sdgPickVertex(x, y);
  else
    sdSubdivTriangle(TM, sdgPickTriangle(x, y));
} /* sdgPickTriangleOrVertex */


static void sdgPickTriangleAndMerge(int x, int y)
{
  if (TM->progressive)
    sdMergeAllLeafTrianglesOnce(TM);
  else
    sdMergeTriangle(TM, sdgPickTriangle(x, y));
} /* sdgPickTriangleAndMerge */


static void sdgPickTriangleAndHighlight(int x, int y)
{
  sdTrianglePtr t;
  int eno;
  sdEdgePtr e;

  t = sdgPickTriangle(x, y);
  if (TM->verbose)
    sdiPrintTriangle(TM, t, USELESS, TRUE);
  sdgHighlightTriangle(t, 1);
  for (eno=0; eno<TM->Nedges; eno++) {
    e = TM->edge + eno;
    e->highlighted = (t && (e->adt[0] == t || e->adt[1] == t));
  }
} /* sdgPickTriangleAndHighlight */


static void sdgRotatePolygonMode(void)
{
  GLenum *pm = sdgPolyMode[DL_MAIN_MESH];

  if (TM->closed) {
    if (pm[0] == GL_FILL)		/* (LINE, LINE) -> (FILL, FILL) */
      pm[0] = pm[1] = GL_LINE;
    else				/* (FILL, FILL) -> (LINE, LINE) */
      pm[0] = pm[1] = GL_FILL;
  } else {
    if (pm[0] == GL_FILL) {
      if (pm[1] == GL_FILL)		/* (FILL, FILL) -> (LINE, LINE) */
	pm[0] = pm[1] = GL_LINE;
      else				/* (FILL, LINE) -> (FILL, FILL) */
	pm[1] = GL_FILL;
    } else {
      if (pm[1] == GL_LINE)		/* (LINE, LINE) -> (LINE, FILL) */
	pm[1] = GL_FILL;
      else {				/* (LINE, FILL) -> (FILL, LINE) */
	pm[0] = GL_FILL;  pm[1] = GL_LINE;
      }
    }
  }
  TM->refresh = TRUE;
} /* sdgRotatePolygonMode */


static void sdgToggleTargetMeshMode(void)
{
  sdgMesh[DL_TGT_MESH]->display = !sdgMesh[DL_TGT_MESH]->display;
} /* sdgToggleTargetMeshMode */


static void sdgToggleCracks(void)
{
  TM->conforming = !TM->conforming;
  TM->refresh = TRUE;
} /* sdgToggleCracks */


static void sdgToggleNormalMode(void)
{
  TM->normalsPerVertex = !TM->normalsPerVertex;
  TM->refresh = TRUE;
} /* sdgToggleNormalMode */


static void sdgToggleDitheringMode(void)
{
  static sdBoolean dithering = TRUE;

  if (dithering = !dithering)
    glEnable(GL_DITHER);
  else
    glDisable(GL_DITHER);
} /* sdgToggleDitheringMode */


static void sdgRotateDetailsLevel(void)
{
  static int level = 0;

  fprintf(stderr, "Subdividing & adding details up to level %d (all bits)...  ",
	  level);
  sdRecalculateAllVertexPositions(TM, 0, level, NBPDN, TRUE);
  fprintf(stderr, "Done.\n");
  if (++level > TM->depth) level = 0;
} /* sdgRotateDetailsLevel */


static void sdgRotateDetailsNbits(void)
{
  static int nbits = 0;

  fprintf(stderr, "Subdividing & adding all details (1+%2d bits)...  ", nbits);
  sdRecalculateAllVertexPositions(TM, 0, TM->depth, nbits+1, TRUE);
  fprintf(stderr, "Done.\n");
  if (++nbits > NBPDN-1) nbits = 0;
} /* sdgRotateDetailsNbits */


static void sdgRotateVerboseMode(void)
{
  if (++TM->verbose > 2) TM->verbose = 0;
} /* sdgToggleVerboseMode */


static void sdgPrintMeshMathematica(void)
{
  sdPrintMesh(TM, MATHEMATICA);
} /* sdgPrintMeshMathematica */


static void sdgPrintMeshVRML2(void)
{
  sdPrintMesh(TM, VRML2);
} /* sdgPrintMeshVRML2 */


static void sdgPrintMeshUseless(void)
{
  sdPrintMesh(TM, USELESS);
} /* sdgPrintMeshUseless */


static void sdgPrintAllDetailTrees(void)
{
  sdiPrintAllDetailTrees(TM);
} /* sdgPrintAllDetailTrees */


static void sdgReadOrWriteDetailsFile(sdBoolean animate)
{
  static sdBoolean firstCall = TRUE;
  int i;

  if (TM->fDetailsIn) {
    if (firstCall && TM->depth < TM->targetDepth) {
      for (i=TM->depth; i<TM->targetDepth-1; i++) {
	sdSubdivAllLeafTrianglesOnce(TM);
	sdgDisplayMesh();
      }
      sdSubdivAllLeafTrianglesOnce(TM);
      firstCall = FALSE;
    }
    sdiReadDetailsFile(TM, animate);
  } else
    sdiWriteDetailsFile(TM);
} /* sdgReadOrWriteDetailsFile */


static void sdgReadOrWriteDetailsFilePlainly(void)
{
  sdgReadOrWriteDetailsFile(FALSE);
} /* sdgReadOrWriteDetailsFilePlainly */


static void sdgReadOrWriteDetailsFileGradually(void)
{
  sdgReadOrWriteDetailsFile(TRUE);
} /* sdgReadOrWriteDetailsFileGradually */


static void sdgPrintButtonsAndKeysMapping(void)
{
  int i;
  static char *help[] = {
    "",
    "BUTTON      ACTION",
    "------      ------",
    "Left        Pick triangle and subdivide it (or, in progressive mode,",
    "            Pick vertex and move it around -- can be undone with U);",
    "Middle      Pick triangle and merge it with its brothers;",
    "Right       Pick triangle and highlight it and its neighbours.",
    "",
    "KEY(S)      ACTION",
    "------      ------",
    "LRUD arrow  Rotate camera Left, Right, Up or Down around object resp.;",
    "468239      Move camera left, right, up, down, forward or backward resp.;",
    "W           Rotate wire-frame/polygon mode;",
    "C           Toggle cracks on/off;",
    "N           Toggle normal mode per facet/vertex;",
    "B           Rotate blablabla (verbose) mode;",
    "I           Toggle dithering on/off;",
    "T           Toggle target mesh on/off (progressive mode only);",
    "s           Subdivide all triangles once;",
    "S           Subdivide until Smoothness is achieved (adaptive mode only);",
    "X           Merge all triangles once;",
    "l           Rotate level of details (progressive mode only);",
    "L           Rotate nr. of bits of details (id.);",
    "U           Undo last vertex move (id.);", 
    "E           Rotate initial mesh edge (id.);",
    "M           Print mesh in Mathematica format;",
    "V           Print mesh in VRML 2.0 format;",
    "p           Print mesh in CU (Completely Useless) format;",
    "P           Print detail trees in CU format (progressive mode only);",
    "d           Read/Write details file (id.);",
    "D           Read/Write details file and animate mesh (id.);",
    "H           Print this help message;",
    "Q, Esc      Quit.",
    "",
    0
  };

  for (i=0; help[i]; i++)
    fprintf(stderr, "%s\n", help[i]);
} /* sdgPrintButtonsAndKeysMapping */


static void sdgButtonClicked(int button, int state, int x, int y)
{
  /* ignore "un-clicks" */
  if (state == GLUT_UP)
    return;

  switch (button) {
    case GLUT_LEFT_BUTTON:	sdgPickTriangleOrVertex(x, y);		break;
    case GLUT_MIDDLE_BUTTON:	sdgPickTriangleAndMerge(x, y);		break;
    case GLUT_RIGHT_BUTTON:	sdgPickTriangleAndHighlight(x, y);	break;
  }
  glutPostRedisplay();
} /* sdgButtonClicked */


#define ESC 0x1B

static void sdgKeyPressed(unsigned char key, int x, int y)
{
  switch (key) {
    case '4':	sdgMoveCameraLeft();			break;
    case '6':	sdgMoveCameraRight();			break;
    case '8':	sdgMoveCameraUp();			break;
    case '2':	sdgMoveCameraDown();			break;
    case '3':	sdgMoveCameraForward();			break;
    case '9':	sdgMoveCameraBackward();		break;
    case 'w':
    case 'W':	sdgRotatePolygonMode();			break;
    case 'c':
    case 'C':	sdgToggleCracks();			break;
    case 'n':
    case 'N':	sdgToggleNormalMode();			break;
    case 'b':
    case 'B':	sdgRotateVerboseMode();			break;
    case 't':
    case 'T':	sdgToggleTargetMeshMode();		break;
    case 'i':
    case 'I':	sdgToggleDitheringMode();		break;
    case 's':	sdgSubdivAllLeafTrianglesOnce();	break;
    case 'S':	sdgSubdivAllLeafTrianglesUntilSmooth();	break;
    case 'x':
    case 'X':	sdgMergeAllLeafTrianglesOnce();		break;
    case 'l':	sdgRotateDetailsLevel();		break;
    case 'L':	sdgRotateDetailsNbits();		break;
    case 'u':
    case 'U':	sdgUndoMoveVertex();			break;
    case 'e':
    case 'E':	sdgHighlightNextEdgeTargetMeshPath();	break;
    case 'm':
    case 'M':	sdgPrintMeshMathematica();		break;
    case 'v':
    case 'V':	sdgPrintMeshVRML2();			break;
    case 'p':	sdgPrintMeshUseless();			break;
    case 'P':	sdgPrintAllDetailTrees();		break;
    case 'd':	sdgReadOrWriteDetailsFilePlainly();	break;
    case 'D':	sdgReadOrWriteDetailsFileGradually();	break;
    case 'h':
    case 'H':	sdgPrintButtonsAndKeysMapping();	break;
    case 'q':
    case 'Q':
    case ESC:	sdgDestroyMeshAndCloseDisplay();	break;
  }
  glutPostRedisplay();
} /* sdgKeyPressed */


static void sdgSpecialKeyPressed(int key, int x, int y)
{
  switch (key) {
    case GLUT_KEY_LEFT:		sdgRotateCameraLeft();	break;
    case GLUT_KEY_RIGHT:	sdgRotateCameraRight();	break;
    case GLUT_KEY_UP:		sdgRotateCameraUp();	break;
    case GLUT_KEY_DOWN:		sdgRotateCameraDown();	break;
  }
  glutPostRedisplay();
} /* sdgSpecialKeyPressed */


void sdDisplayMeshForever(sdMeshPtr m, int width, int height, char *title)
{
  if ((sdgMesh[DL_MAIN_MESH] = m))
    m->display = m->refresh = TRUE;
  else
    return;

  if ((sdgMesh[DL_TGT_MESH] = m->targetMesh))
    m->targetMesh->display = m->targetMesh->refresh = TRUE;
  sdgInitDisplay(width, height, title);
  glutMouseFunc(sdgButtonClicked);
  glutMotionFunc(sdgMoveVertex);
  glutKeyboardFunc(sdgKeyPressed);
  glutSpecialFunc(sdgSpecialKeyPressed);
  sdgPrintButtonsAndKeysMapping();
  glutDisplayFunc(sdgDisplayMesh);
  glutMainLoop();
} /* sdDisplayMeshForever */

